"""

Bagua_ is a communication library
developed by Kuaishou Technology and DS3 Lab for deep learning.

See tutorials_ for Bagua's rationale and benchmark.

.. _Bagua: https://github.com/BaguaSys/bagua
.. _tutorials: https://baguasys.github.io/tutorials/
"""

import bagua_core  # noqa: F401
from .version import __version__  # noqa: F401
